;***************************************************************************
;* Title: AVR_POCSAG pocsag encoder test code Rev. 1.1
;* Made available for educational, and non-commercial use.
;* Commercial use forbidden without license from author.
;* Code may be modified as long as it is released publicly and credit
;* is given to the original author(s).  
;* (c) copyright 2000 Henry Carl Ott N2RVQ, All Rights Reserved.
;* Questions, comments or offers of cash? carlott@rcn.net  http://users.rcn.com/carlott/
;***************************************************************************
;* Description: A standalone pocsag encoder. Unit will accept strings from console and encode
;* as standard POCSAG. Alphanumeric, numeric only, and tone messages can be encoded at 
;* 512/1200/2400 baud and as inverted or non-inverted data. In addition support is provided for
;* transmission of Morse encoded strings at varying speed rates, radio control, and certain
;* debugging functions. Please see companion documentation for schematics, use, and command syntax.  
;****************************************************************************
;* Credits: The asc2bin routine is a modified/limited version of a routine by Jack Tidwell as made
;* available in his avrmath.asm code. Credit is also given to everyone who made example AVR
;* code available on the net for me to learn how to program these Atmel processors.
;***************************************************************************
;* Disclaimer: This code is provided "as is".  Don't use this code/design for any
;* application where human life is at risk  or property damage may occur.
;***************************************************************************  
;Versions:
; 1.1 fixed timing error in isr timer reload value
; 1.0 first release
 
.nolist
.include "2313def.inc"	
.list
.listmac

;***************************************************************************
;*
;* MACROS
;*
;***************************************************************************
;--- simple led macro
.macro	led_on
	sbi	PORTD,led
.endm	
.macro	led_off
	cbi	PORTD,led
.endm	

;--- send single char to console
.macro	_send
	ldi	temp,@0 
	rcall	putc
.endm	

;--- send string to console
.macro _send_232_str
	ldi	zh,high(@0 * 2)
       	ldi	zl,low(@0 * 2) 	;Init Z-pointer
	rcall	ser_ram_str
.endm

;--- copy flash data to pocsag output buffer
.macro	_flash2buff
	ldi	zh,high(@0 * 2)
       	ldi	zl,low(@0 * 2) 	;Init Z-pointer
	rcall	flash2buff
.endm

;--- skip next instruction
.macro	skip
	rjmp	(pc + 2)	; skip, next instruction	
.endm	

;------delay two cycles in one instruction
.macro	_nop_2
	rjmp	(pc + 1)
.endm	



;--- some program equates and constants
.equ	ram_begin	= 0x60		; start of ram on 2313	
.equ	max_str_len	= 80		; max len of string from console
.equ	getc_to_val	= 4		; time before getc will timeout (in approx 2.5 second incs)
.equ	eom_char	= eot		; char to append to alpha messages (may not be necessary)	
.equ	clock 		= 7372800 	; cpu clock
.equ	console_baud 	= 9600
.equ	isr_divider 	= (0xffff - (clock/9600)) + 15

;--- ascii char codes
.equ	ctrl_c 		= 0x03
.equ	etx		= 0x03
.equ	eot		= 0x04
.equ	bel 		= 0x07 		; bell
.equ	bs  		= 0x08		; back space
.equ	cr  		= 0x0d		; carriage return 
.equ	lf  		= 0x0a		; line feed


;.equ	frame_sync	= 0x7CD215D8	;  pocsag frame sync word (for reference)
;.equ	idle 		= 0x7A89C197	;  pocsag idle word 	  (for reference)

;-----------------------------------------------------------------
;--- port and pin defines
;--- portb 
.equ spare_0	= 0
.equ cfg_0	= 1	; unused at moment	
.equ cfg_1	= 2	; unused at moment 
.equ cw_out	= 3	; tone output for cw generation
.equ poc_out	= 4	; encoded pocsac data out
.equ inp_1	= 5	; send canned message 1
.equ inp_2	= 6	; send canned message 2	
.equ tx_inhibit	= 7	; x-mitter lockout

;--- portd
.equ ptt	= 6 	; key radio
.equ led	= 5	; status
.equ cts_232 	= 4	; clear to send output, asserts while transmitting to indicate busy 
.equ spare_1    = 3	;
.equ spare_2	= 2	;  
.equ txd_232	= 1	; O 232_txd 
.equ rxd_232	= 0	; I 232_rxd

;*****************************
;* Global Register Variables *
;*****************************
.def	r0		= r00	; used with lpm instruction
.def	tic		= r01	; incs @ 9600 hz resets at 96 0-95
.def	sec_01		= r02	; incs @ 100 hz	
.def	u_data  	= r03	; rs-232 data (rx)
.def	EEaddr		= r04	; internal eeprom address buffer
.def	EEdata		= r05	; internal eeprom data buffer for rd/wr 
.def	frame		= r06	; current pocsag frame 0-15

.def	mant1hh		= r07	
.def	mant1h		= r08	; used for ascii to bin conversion
.def	mant1m		= r09	; also mapped to bin
.def	mant1		= r10

.equ	at_bin		=  07	; pointer location of bin registers	
.def	binhh		= r07	; msb	
.def	binh		= r08	; working registers to build pocsag word
.def	binm		= r09
.def	bin		= r10	

.def	mant2hh		= r11
.def	mant2h		= r12	; used for asc2bin routines 
.def	mant2m		= r13	; also mapped to pocbuff
.def	mant2		= r14

.equ	at_poc_buff	=  11	; pointer location of buffer regs 
.def	poc_buff_hh	= r11	; 32 bits of data to be clocked out by isr
.def	poc_buff_h	= r12
.def	poc_buff_m	= r13
.def	poc_buff	= r14
.def	isr_save	= r15	; saves mcu status reg during isr 

.def	temp		= r16	; gp working reg 	
.def	temp2		= r17   ; yet another working register 
.def	ii		= r18	; gp loop counter
.def	flags		= r19	; various bit flags for isr
.def	poc_flags	= r20	; various bit flags of message being sent	
.def	byte_cnt	= r21	; general purpose byte counter
.def	bit_cnt		= r22	; gp bit counter
.def	morse_rate	= r23	; dit length in .01 incs 	 
.def	poc_bit_cnt	= r24	; bit counter for isr x-mit routine 1-32
.def	poc_baud_div	= r25	; divide isr 9600 tic to 2400,1200,512

.def	poc_char_pnt	= r26	; pointer to next message char to process 
.def	poc_char	= r27	; current char being processed
.def	yl		= r28	; used in flash2buff to point to destination
.def	yh		= r29	; 
.def	zl		= r30	; used for indirect indexing   
.def	zh		= r31	; used for indirect indexing 

;**************************
;- isr bit flags defined
;**************************
.equ	p_baud_1200	= 0x00		; pocsag baud rate for isr  
.equ	p_baud_2400	= 0x01		; ditto, if neither bit is set, then baud = 512
.equ	p_invert	= 0x02		; invert pocsag data before sending
.equ	cw_tone		= 0x03		; turns on tone generation in isr
.equ	uart_rx		= 0x04		; incoming char is ready
.equ	debug		= 0x05		; if set, echo pocsag data to console
.equ	poc_pre0	= 0x06		; divide by 4 prescaler for pocsag  512 baud
.equ	poc_pre1	= 0x07		; ditto

;**************************
;- pocsag stored message bit flags defined
;**************************
.equ	m_baud_1200	= 0x00	
.equ	m_baud_2400	= 0x01		; if neither bit, then 512
.equ	m_invert	= 0x02		; invert?
.equ	m_spare		= 0x03		; not used
.equ	m_fun_0		= 0x04		; msg function
.equ	m_fun_1		= 0x05		; msg function
.equ	m_type_num	= 0x06		; msg type numeric
.equ	m_type_alpha	= 0x07		; msg type alpha

;-------------------------

;*********************
;* Interrupt Vectors *
;*********************

.CSEG
.org		0x00				; reset 
		rjmp	reset			
.org		INT0addr			; external int 0 	
		reti
.org		INT1addr			; external int 1
		reti
.org		OVF1addr			; overflow of 16 bit timer 1 
		reti			
.org		OVF0addr			; overflow of 8  bit timer 0 ; 9600hz system tic 
		rjmp	t0int
.org 		URXCaddr			; reception of data by uart
		rjmp	uart_rx_isr		

;--- timer 1 overflow interupt, we should hit the isr @ 9600 hz		
;--- the isr shifts out the 32 bit pocsag word at the correct baud rate
;--- isr will also generats cw tone, and incs the sec_001 counter @ 100 hz
t0int:		in	isr_save,SREG		; save SREG
		push	temp	
		
		ldi	temp,256 - ((clock/8) / 9600) + 1 ;(1 = approx overhead)
		out	TCNT0,temp		; 

		inc	poc_baud_div		; 
		inc	tic			; bump counter everytime through
		ldi	temp,96			; divide by 96 to get our 10ms tic	
		eor	temp,tic		; 
		brne	no_roll_over		; no rollover		

		clr	tic
		inc	sec_01			; bump hunds of seconds (10ms)

no_roll_over:	sbrs	flags,cw_tone		; is tone enabled ?
		rjmp	sel_poc_baud		; no

		sbrc	tic,3			; 1200 hz out
		sbi	portb,cw_out		
		sbrs	tic,3
		cbi	portb,cw_out	
		rjmp	tmr0isr_end		; skip pocsag generation		

sel_poc_baud:	sbrc	flags,p_baud_1200	; test baud rate bits
		rjmp	set_1200		;
		sbrc 	flags,p_baud_2400	;
		rjmp	set_2400	
					
set_512:	mov	temp,flags		; (9600*4) / ((3*19)+18)) = 512
		andi	temp,0b11000000		; so, 1 in 4 times we count to 18 
		breq	div18			; other three times count to 19
div19:		ldi	temp,1
div18:		subi	temp, -18		; add 18 
		cp 	poc_baud_div,temp	; 
		brne 	tmr0isr_end		; no match

		ldi	temp,0x40		; matched	
		add	flags,temp		; top two bits of flag register is used to divide by 4
		rjmp	period			; process bit

set_1200:	cpi 	poc_baud_div,8		; 1200 = 9600 / 8
		breq 	period			; match	
		rjmp 	tmr0isr_end		; no match

set_2400:	cpi	poc_baud_div,4		; 2400 = 960 /4
		brne	tmr0isr_end		; no match
	
period:		clr	poc_baud_div		; reset baud rate divisor
		tst	poc_bit_cnt		; test for zero bits
		breq	tmr0isr_end		; nothing to send
	
		ldi	temp,0x80	 	; if invert flag is set, we invert the bit prior to sending
		sbrc	flags,p_invert		; 
		eor	poc_buff_hh,temp		

		sbrs	poc_buff_hh,7		; copy hi bit to output pin
		cbi	portb,poc_out		; 
		sbrc 	poc_buff_hh,7		; double test for no jitter
		sbi	portb,poc_out		; 	
		
		rol	poc_buff		; rotate next bit into position
		rol	poc_buff_m
		rol	poc_buff_h
		rol	poc_buff_hh		; msb

		dec	poc_bit_cnt		; dec number of bits left to send

tmr0isr_end:	pop	temp			; 
		out	SREG,isr_save		; restore
		reti

;--- rs-232 char reception isr, checks for errors and discards bad chars  
uart_rx_isr:	in	isr_save,SREG		; store SREG
		push	temp			
		sbic	USR, FE			; skip if framing error
		rjmp	uart_err		; 
		in	temp, UDR		; get rx byte
		sbic	USR, OR			; skip if overrun error
		rjmp	uart_err			; 
		mov	u_data,temp		; save actual data
		sbr	flags,(1<<uart_rx)	; set received data flag
		rjmp	uart_isr_end

uart_err:	in	temp, UDR		; read and discard byte to clear UDR
		cbr	flags,(1<<uart_rx)	; clear flag return

uart_isr_end:	pop	temp			; restore
		out	SREG,isr_save		; restore	
		reti

;------------------------------------------------------------------------
;--- start here on power up or external reset
;--- init ports, ints,  		
;--- zero all  registers, maybe not necessary but a good idea while debugging 		
reset:		ldi	r16,0
		ldi	zl,0				; beginning of registers		
clr_reg:	st	z+,r16
		cpi	zl,0x1e				; don't zero z reg
		brne	clr_reg

;--- init stack pointer (somewhat important)
		ldi    temp,low(RAMEND)			; init stack 
       		out    spl,temp 
;--- portb setup
		ldi	temp,(1 << cw_out)|(1 << poc_out)
		out	DDRB,temp
		sbi	PORTB,cfg_0			; pull up	
		sbi	PORTB,cfg_1			; pull up
;--- portd setup
		ldi	temp,(1 << led)|(1 << cts_232)|(1 << txd_232)|(1 << ptt)
		out	DDRD,temp
;--- setup uart, 9600 baud, int on rx, no tx ints
		ldi	temp,((clock / console_baud) / 16) -1
		out	UBRR,temp			; Set baud rate generator
		ldi	temp, (1<<RXCIE)|(1<<RXEN)|(1<<TXEN) ; enable rxc interrupts
		out	UCR,temp			; enable UART tx & rx  w/o interrupts
;--- set up tmr0

		ldi	temp,2				; timer 0 prescale/8
		out	TCCR0,temp			;

		ldi	temp,0b00000010			; enable Timer 0 interrupt
		out	TIMSK,temp

;--- global enable ints
		sei					; enable global ints
		ldi	morse_rate,0x05			; default morse wpm rate	
		
		_send_232_str	_hello			; display sign on message

		led_on					; flash the led
		ldi	temp,100
		rcall	del_hunds
		led_off
		
		rjmp	cmd_top				; wait for command

error:		_send_232_str	_error
;----------------------------------------------
cmd_top:	cbi	PORTD,cts_232			; enable comm
		rcall	crlf				; prompt
		_send	'.'


inp_loop:	sbis	PINB,inp_1			; inp 1 or 2 down?
		rjmp	send_inp_1		
		sbis	PINB,inp_2	
		rjmp	send_inp_2
		sbrs	flags,uart_rx			; any console data?
		rjmp	inp_loop

	   	rcall	get_string
		brcc	pc+2				; valid string?
		rjmp	cmd_top

		ldi	zl,ram_begin			; convert first char to upper case
		ld	temp,z
		rcall	to_upper

do_cmd:		cpi	temp,'D'			; set debug mode
		brne	pc+2
		rjmp	set_debug

		cpi	temp,'I'			; test inhibit line
		brne	pc+2				; return status
		rjmp	check_inhibit

		cpi	temp,'M'			; send morse string	
		brne	pc+2
		rjmp	send_morse

		cpi	temp,'R'			; get morse wpm
		brne	pc+2
		rjmp	set_m_rate

		cpi	temp,'G'			; get a message and store in eeprom
		brne	pc+2
		rjmp	get_eemsg

		cpi	temp,'S'			; send canned message from eeprom
		brne	pc+2
		rjmp	send_eemsg
	
		cpi	temp,'P'			; send pocsag message directly 
		breq	send_con_str
		rjmp	cmd_top

;---------------------------------------------------------------------------------
;--- set and display debug mode
set_debug:	lds	temp,ram_begin+1				; get arg
		cpi	temp,'0'
		brne	pc + 2
		cbr	flags, 1 << debug		; set appropriate flag
		cpi	temp,'1'
		brne	pc + 2
		sbr	flags, 1 << debug
		_send_232_str	_debug			; message
		ldi	temp,'0'
		sbrc	flags,debug
		ldi	temp,'1'
		rcall	putc				; send status
		rjmp	cmd_top

;--- test and return status of tx_inhibit line
check_inhibit:	rcall	crlf
		ldi	temp,'0'
		sbic	PINB,tx_inhibit
		ldi	temp,'1'
		rcall	putc
		rjmp	cmd_top	

;--- send console string via morse	
send_morse:	sbi	PORTD,cts_232
		ldi	ii,1
		rcall	pos_zl
		mov	xl,zl				; point to data

		rcall	key_radio
		rcall	send_cw_string
		rcall	unkey_radio
		rjmp	cmd_top	

;--- set morse wpm rate
set_m_rate:	ldi	ii,1
		rcall	pos_zl
		rcall	asc2bin
		mov	morse_rate,bin
		rjmp	cmd_top

;--- send console string as pocsag directly 
send_con_str:	sbi	PORTD,cts_232			; disable comm
		rcall	count_param
		cpi	byte_cnt,6			; check for minimum num of args	
		brsh	pc + 2	
		rjmp	error
		rcall	parse_msg			; parse out cap code and flags

		rcall	test_cap			; make sure cap code is within valid range
		brcc	pc +2
		rjmp	error

		ldi	ii,6
		rcall	pos_zl				; get pointer to message data	
		mov	poc_char_pnt,zl
		rcall	send_page			; send it
		rjmp	cmd_top				

;--- parse console string  and store in one of two eeprom storage areas
get_eemsg:	sbi	PORTD,cts_232			; disable comm
		rcall	count_param
		cpi	byte_cnt,6			; check for minimum num of args	
		brsh	pc + 2
		rjmp	error				; not enough args	

		rcall	set_eeaddr			; get second char of string, set EEaddr to msg1_pin or msg2_pin
		breq	pc + 2
		rjmp	error
	
		rcall	parse_msg			; get pin and flags	
		rcall	test_cap			; is pin in valid range?
		brcc	pc +2
		rjmp	error

		mov	EEdata,binh			; save cap code to eprom
		rcall	EEwrite
		mov	EEdata,binm
		rcall	EEwrite
		mov	EEdata,bin
		rcall	EEwrite
		mov	EEdata,poc_flags
		rcall	EEwrite				; save the flag bits

		ldi	ii,6				; now get message data		
		rcall	pos_zl				; point to data
parse_dl:	clr	EEdata	
  		mov	temp,EEaddr
 		andi	temp,0b00111111			; only 64 chars allowed per message
 		cpi	temp,0b00111111		
 		breq	pc + 2	
		ld	EEdata,z+			; get char
		rcall	EEwrite				; save char
		tst	EEdata				; tst for zero termination
		brne	parse_dl			; get next char	
		rjmp	cmd_top
		
;--- send a canned message from eeprom
send_eemsg:	sbi	PORTD,cts_232			; disable comm
		rcall	set_eeaddr
		breq	pc + 2
		rjmp	error
		rcall	send_msg_ee
		rjmp	cmd_top				


;--- get second char of string
;--- load EEaddr, z is cleared if no match (bad)
set_eeaddr:	lds	temp2,ram_begin + 1		; get second char of string	
		ldi	temp,msg1_pin
		cpi	temp2,'1'
		breq	set_ee_end			; if '1' send message 1	
		ldi	temp,msg2_pin
		cpi	temp2,'2'
		breq	set_ee_end			; and etc.	
		clz					; no match, back to top
set_ee_end:	mov	EEaddr,temp
		ret


;-------------------------------------------------------------
;--- wait for hw input to de-assert, then send canned message 1 or 2
send_inp_1:	sbi	PORTD,cts_232			; disable comm
		ldi	temp,msg1_pin			; load pointer	
		clr	temp2
		led_on					; status
s_inp_1_lp:	sbis	PINB,inp_1			; debounce
		clr	temp2				; reset
		dec	temp2
		breq	send_inp_x			; do_it
		rjmp	s_inp_1_lp			; loop
;---
send_inp_2:	sbi	PORTD,cts_232
		ldi	temp,msg2_pin
		clr	temp2
		led_on
s_inp_2_lp:	sbis	PINB,inp_2			; debounce
		clr	temp2				; reset
		dec	temp2
		breq	send_inp_x			; do_it
		rjmp	s_inp_2_lp			; loop

send_inp_x:	mov	EEaddr,temp
		rcall	send_msg_ee
		rjmp	cmd_top	


;--- send message from eprom
;--- call with EEaddr pointing to pin
send_msg_ee:	rcall	get_ee_msg			; copy eeprom data to regs and ram
		ldi	poc_char_pnt,ram_begin		; point to data

;------------------------------------------------------------------
;--- send a complete page from ram
;--- bin should have cap code
;--- poc_flags should be loaded
;--- poc_char_ptr points to first char of data
send_page:	mov	byte_cnt,bin
		andi	byte_cnt,0x07			; get low three bits of address and save 
		lsl	byte_cnt			; multiply by 2

		mov	temp,bin			
		andi	temp,0b11111000			; mask off low bits
		sbrc	poc_flags,m_fun_1
		sbr	temp, 1 << 2			; copy in function bits 
		sbrc	poc_flags,m_fun_0
		sbr	temp, 1 << 1			; copy in function bits 
		mov	bin,temp			; copy it back
		
		;clr	binhh				; clear any garbage
		lsr	binh
		ror	binm
		ror	bin				; move to final bit positions
		rcall	ecc_bin				; add ecc and parity 

		mov	temp,poc_flags			; copy baud and invert status from stored message to isr flags
		andi	temp,0b00000111			; mask off high bits
		andi	flags,0b11111000		; mask off low bits
		or	flags,temp			; now mask in low bits

		rcall	key_radio			; key radio
		ldi	temp,50
		rcall	del_hunds			; wait a half second

		ldi	ii,18
send_preamble:	_flash2buff	_0xaa			; send preamble, 576 bit reversals				
		dec	ii
		brne	send_preamble
		rcall	new_frame			; new frame, send sync, clear frame counter	

wait4frame:	cp	frame,byte_cnt			; match?
		breq	send_message
		_flash2buff	_idle			; no match, send idle, keep looping	
		rcall	next_frame
		rjmp	wait4frame			; loop it

send_message:	rcall	bin2buff			; send address
		rcall	next_frame			; bump frame counter

		mov	temp,poc_flags
		andi	temp,0b11000000			; tone only message?
		breq	send_p_end			; then all done

		clt					; init for first data char
		ldi	bit_cnt,20			; 20 bits to pack per frame 
		ldi	byte_cnt,0			; count bits within char (4 bits tone) (7 bits alpha) 	

send_data_lp:	rcall	pack_data
		brcs	send_p_end			; if carry is set, there is no data to send
		rcall	bin2buff			; send packed data
		rcall	next_frame			; bump frame counter add frame sync as necessary
		rjmp	send_data_lp			; loop for more data

send_p_end:	rcall	fill_batch			; fill batch in progress with idle frames
wait_p_done:	tst	poc_bit_cnt			; wait for tx buffer to empty out
		brne	wait_p_done
		rcall	unkey_radio			; all done
		ret

;------- pocsag xmit support subroutines
;--- fill current batch with idles	
fill_batch:	sbrc	frame,4				;  are we at 16?		
		ret					; yes, all done
		_flash2buff	_idle			; send a idle frame	
		inc	frame				
		rjmp	fill_batch			; keep looping
;--- inc frame counter. if necessary send a frame sync and reset frame counter
next_frame:	inc	frame				; bump frame counter
		sbrs	frame,4				; roll over at 16
		ret
new_frame:	_flash2buff	_frame_sync		; send a frame sync char	
		clr	frame				; clear frame counter
		ret

;--- get chars from ram, pack them into 20 bit frames, ecc the data and return
;--- if no more data, set carry
;--- enter with xl pointing to first byte of zero terminated data string
;--- uses temp,bit_cnt,byte_cnt,poc_char,poc_char_ptr
;--- t bit flag indicates padding in progress
pack_data:	tst	byte_cnt			; any bits from left from last char?
		breq	pc + 2				; nope
		rcall	stuff_char			; otherwise shift in those left over bits	
		brts	pad_data			; last frame was padded, finish up

get_poc_char:	ld	poc_char,x+			; get char
		tst	poc_char			; all done
		breq	pad_data

		ldi	byte_cnt,7
		sbrc	poc_flags,m_type_num
		rcall	xlate_numbers			; if message numeric traslate ascii, and set byte_cnt to 4

		rcall	stuff_char
		cpi	bit_cnt,20			; did we fill a frame?
		brne	get_poc_char			; nope get another char
		clc					; clear carry to indicate okay data
		ret					; yep, return with encoded data

;--- fill empty bit positions in incomplete data words with appropriate data
;--- numeric message are filled with space charsif necessaary, alphas get at least one eom char
pad_data:	sbrc	poc_flags,m_type_alpha		; letters / numbers
		rjmp	pad_alpha			; 

;---
pad_num:	cpi	bit_cnt,20			; are we at beginning of next frame?	
		brne	pad_num_l			; no, have to pad
		sec					; yes, no data to send
		ret					; and out

pad_num_l:	ldi	poc_char,0x0c
		ldi	byte_cnt,4			;
		rcall	stuff_char			
		cpi	bit_cnt,20			; done ?
		brne	pad_num_l			; keep looping
pad_in_dat:	set					; set t flag to indicate padding was done
		clc					; there is data to send		
		ret

;--- we stick at least one eom char at end of alpha message
pad_alpha:	brtc	pad_alpha_1			; was frame was padded

		cpi	bit_cnt,20			; are we at beginning of next frame?	
		brne	fill_alpha			; no, we have to pad
		sec					; yes, done, no data to send
		ret	

pad_alpha_1:	rcall	stuff_eom			;
		breq	pad_in_dat			; no, send what we have, finish next time through

fill_alpha:	rcall	stuff_eom
		brne	fill_alpha			; 
		clr	byte_cnt			; discard rest of bits
		rjmp	pad_in_dat			; set status and return

;--- stuff eom char, test for full buffer
stuff_eom:	ldi	poc_char,eom_char		; fill rest of bin with eoms			
		ldi	byte_cnt,7			; 7 bits per char
		rcall	stuff_char
		cpi	bit_cnt,20
		ret

;--- rotate lsb of char being processed into lsb of bin
;--- dec byte_cnt, if more bits keep stuffing, otherwise exit
;--- when 20 bits are done, set high bit to indicate a data frame
;--- ecc bin, clear bit_cnt, and return
stuff_char:	ror	poc_char			; rotate lsb into carry

		rol	bin				; rotor it out
		rol	binm				; lsb of raw data is msb of frame
		rol	binh

		dec	bit_cnt				; amy more room in frame?
		breq	stuff_full			; nope	

		dec	byte_cnt			; all bits in char processed?
		brne	stuff_char			
		ret			

stuff_full:	dec	byte_cnt			; we have to count that last bit	
		ldi	temp,0b00001111
		and	binh,temp			; clear out any garbage
		ldi	temp,0b00010000
		or	binh,temp			; set high bit to indicate data
		rcall	ecc_bin				; add ecc and parity	
		ldi	bit_cnt,20 			; set up for next data frame
stuff_end:	ret 



;---  translate ascii data into numeric data
;--- numbers convert to numbers (duh), spaces, hyphens, brackets are translated individually
;--- non valid characters will convert to spaces 
xlate_numbers:	ldi	byte_cnt,4		; four bits for char stuff routine
		cpi	poc_char,'9'+1
		brsh	xlate_non_num
		cpi	poc_char,'0'
		brlo	xlate_non_num
		subi	poc_char,'0'
		ret				; 0-9, just convret to bcd  and return

xlate_non_num:	mov	temp,poc_char		; save it
		rcall	to_upper
		ldi	poc_char,0x0c		; preload with space char

		cpi	temp,'U'		; urgent char	
		brne	pc + 2
		ldi	poc_char,0x0B	
		cpi	temp,'-'
		brne	pc + 2
		ldi	poc_char,0x0D	
		cpi	temp,']'
		brne	pc + 2
		ldi	poc_char,0x0E
		cpi	temp,'['
		brne	pc + 2
		ldi	poc_char,0x0F
		ret

;--- parse string,  get pin,function,baud rate,invert status,msg type, 
;--- pin goes to bin, message flags goto poc_flags, leaves data in ram 
parse_msg:	ldi	ii,1				; get pin
		rcall	pos_zl
		rcall	asc2bin				; convert to binary
	
		ldi	ii,2				; now get msg options and mask into temp2
		rcall	pos_zl			

		ld	poc_flags,z
		andi	poc_flags,0x03			; mask off high bits
		swap	poc_flags			; move bits into location
		
		ldi	ii,3				; get baud
		rcall	pos_zl				; 2400,1200,512
		ld	temp,z
		cpi	temp,'2'			; if no match, will default to 512
		brne	pc + 2
		sbr	poc_flags, 1 << m_baud_2400	
		cpi	temp,'1'
		brne	pc + 2
		sbr	poc_flags, 1 << m_baud_1200	

		ldi	ii,4				; get message polarity, invert if 'I' 
		rcall	pos_zl				; else normal
		ld	temp,z
		rcall	to_upper
		cpi	temp,'I'
		brne	pc + 2
		sbr	poc_flags, 1 << m_invert
		
		ldi	ii,5				; get mesage type, A,N,T
		rcall	pos_zl				; default to Tone only
		ld	temp,z
		rcall	to_upper
		
		cpi	temp,'A'
		brne	pc + 2
		sbr	poc_flags, 1 << m_type_alpha	
		cpi	temp,'N'
		brne	pc + 2
		sbr	poc_flags, 1 << m_type_num	
		ret		
;---
key_radio:	led_on					; status
		sbis	PINB,tx_inhibit			; wait for inhibit line to de-assert 
		rjmp	key_radio
		led_off
		sbi	PORTD,ptt
		ret

unkey_radio:	cbi	PORTD,ptt
		ret

; make sure cap code (pin) is within  valid range 8 - 2097151
test_cap:	tst	binhh
		brne	test_p_err			 
		mov	temp,binh
		andi	temp,0b11100000			; mask off bottom bits
		tst	temp
		brne	test_p_err			; too big

		tst	binh
		brne	test_p_ok		
		tst	binm
		brne	test_p_ok

		mov	temp,bin
		cpi	temp,8
		brlo	test_p_err			; too small	

test_p_ok:	clc
		rjmp	pc + 2
test_p_err:	sec
		ret

;--- add ecc and even parity to 21 bits of data in bin0 - bin3
;--- bin3 is lsb
; trashes ii,temp,temp2
ecc_bin:	push	bin
		push	binm
		push	binh

		ldi	ii,21		; 21 bits to test
		rjmp	ecc_b		; first bit,  skip initial rotate

ecc_a:		rcall	rol_mant	; next bit
ecc_b:		sbrs	binh,4		; test hi bit of data
		rjmp	ecc_c		; no high bit, skip xor 

		ldi	temp,0x1d	; if the ms bit = 1 the xor in poly	
		eor	binh,temp	
		ldi	temp,0xa4
		eor	binm,temp

ecc_c:		dec	ii		; any more bits?	
		brne	ecc_a		; nope, keep going

		mov	temp,binm	; save ecc data	
		mov	temp2,binh
		lsr	temp2		; shift right to line up data
		ror	temp		; into final bit positions	

		pop	binhh		; get orig data, shifted by one byte	
		pop	binh		; ditto
		pop	binm		; ditto
		
		ldi	ii,3
ecc_dl:		rcall	rol_mant	; left shift into final positions
		dec	ii
		brne	ecc_dl

		mov	bin,temp	; get low bits of ecc
		or	binm,temp2	; or in hi bits of ecc

		mov	temp,binhh	; calc even parity
		eor	temp,binh
		eor	temp,binm
		eor	temp,bin	; reduce 32 bits to 8	

		mov	temp2,temp	; save it, two copies

		swap	temp2		; xor top nibble with lower
		eor	temp,temp2	; 8 bits reduced to 4
		mov	temp2,temp	; save it, two copies

		lsr	temp2
		lsr	temp2		; rotate 2 bit pos
		eor	temp,temp2	; 4 bits reduced to two bits

		mov	temp2,temp	; save it 
		lsr	temp2
		eor	temp,temp2	; parity left in bottom four bits 0000=even 1111=odd
		andi	temp,0x01	; leave only bottom bit
		or	bin,temp	; and mask it in
		ret	

;--- send zero terminated string from ram as morse code
;--- xl points to first char
send_cw_string:	ld	poc_char,x+		; get char
		tst	poc_char		; are we done?
		brne	pc + 2
		ret
		rcall	morse_xlate		; xlate to dits and daahs
		rcall	send_cw_char		; send char		
		rjmp	send_cw_string		; get next char
	
;--- ship single pre xlated char
send_cw_char:	tst	poc_char		; is space char?
		brne	send_char_l		; nope, send it
		rcall	del_word		; 
		ret

send_char_l:	sbrs	poc_char,7		; test high bit, call appropriate routine
		rcall	dit
		sbrc	poc_char,7		; call appropriate routine
		rcall	dah

		lsl	poc_char
		cpi	poc_char,0b10000000
		breq	pc +2
		rjmp	send_char_l		; keep looping
		rcall	del_dah			; inter char delay
		ret
;---
dit:		sbr	flags,(1 << cw_tone)
		rcall	del_dit
		rjmp	dit_dah_end
		
dah:		sbr	flags,(1 << cw_tone)
		rcall	del_dah

dit_dah_end:	cbr	flags,(1 << cw_tone)
		rcall	del_dit
		ret

del_word:	rcall	del_dah			; dit times 6
		
del_dah:	rcall	del_dit			; dit times three		
		rcall	del_dit

del_dit:	clr	sec_01
		mov	temp,morse_rate			;
		rcall	del_hunds
		ret

;--- translate ascii to morse encoded data table
;--- 0=dit 1=dah data is clocked out from the left and terminated by a 1
morse_xlate:	ldi	zl,low(2 * morse_table) 	;Init Z-pointer
		ldi	zh,high(2 * morse_table)

		mov	temp,poc_char
		rcall	to_upper

		subi	temp,0x20		; subtract unused lower 32 chars
		andi	temp,0b00111111		; mask off any upper bits, keep in range 
	
		lsl	temp			; temp * 2 to convert to byte counter
		add	zl,temp
		clr	temp
		adc	zh,temp			; handle carry
			
		lpm
		mov	poc_char,r0
		ret

;--- multiply mant1 by decimal 10
mul_10: 	mov	mant2,mant1		; save mant1
		mov	mant2m,mant1m
		mov	mant2h,mant1h
	
		rcall	rol_mant	; * 2
		rcall	rol_mant	; * 4

		add	mant1,mant2	; * 5
		adc	mant1m,mant2m
		adc	mant1h,mant2h		; fall through to final multiply

;--- rotate mant1 left by one position 
rol_mant:
		lsl	mant1
		rol	mant1m
		rol	mant1h
		rol	mant1hh
		ret

;--- convert ascii string to int (24 bits max)
;--- enter with z pointing to string
asc2bin:	clr	mant1		; start at zero
		clr	mant1m
		clr	mant1h
		clr	mant1hh		; not used

a2b_l:		ld	temp,z+		; get char
		cpi	temp,'9'+1
		brsh	a2b_end
		cpi	temp,'0'
		brlo	a2b_end
		rcall	mul_10
		subi	temp,'0'
		add	mant1,temp	; add new digit
		clr	temp		; just add 0+Carry
		adc	mant1m,temp
		adc	mant1h,temp
		rjmp	a2b_l
a2b_end:	ret


;--- get cr terminated string from console, store as zero terminated string in ram
;--- if terminated by ctrl-c, zero length, set carry
;--- try to support back space, may not work on all terminals
get_string:	ldi	zl,ram_begin

get_str_loop:	rcall	getc
		cpi	temp,cr
		breq	get_str_done

		cpi	temp,ctrl_c
		breq	get_str_abort

		cpi	temp,bs
		brne	save_str_char

		_send	' '
		_send	bs
		dec	zl
		cpi	zl,ram_begin		; we don't want to be able to dec past beginning of ram
		brsh	pc + 2
		ldi	zl,ram_begin
		rjmp	get_str_loop


save_str_char:	cpi	zl,ram_begin + max_str_len
		breq	get_str_loop
		st	z+,temp
		rjmp	get_str_loop
				
get_str_abort:	ldi	zl,0
		sec

get_str_done:	ldi	temp,0
		st 	z,temp
		ret	

;--- set zl to first char of arg, ii
pos_zl:		ldi	zl,ram_begin		; point to beginning
pos_zl_loop:	rcall	next_arg	
		dec	ii
		brne	pos_zl_loop
		ret

;--- count total number of parameters in string, parameters are seperated by ' ' or ','
count_param:	clr	byte_cnt
		ldi	zl,ram_begin				; point to beginning of string	

count_p_loop:	rcall	next_arg
		ld	temp,z
		brne	pc + 2
		ret
		inc	byte_cnt
		rjmp	count_p_loop
	
; inc zl till a non-white char or ','  is encountered
discard_white:	ld	temp,z
		cpi	temp,' '
		breq	discard_loop
		cpi	temp,','
		breq	discard_loop
		ret
discard_loop:	inc	zl					; point to to next location
		rjmp	discard_white

		
; inc zl till a ' ',',', or end of string
discard_nw:	ld	temp,z
		tst	temp					; test for zero termination
		breq	discard_end
		cpi	temp,','
		breq	discard_end	
		cpi	temp,' '
		breq	discard_end
		inc	zl
		rjmp	discard_nw
discard_end:	ret


; position zl at first char of next argument
next_arg:	ld	temp,z
		tst	temp
		brne	pc + 2					; test for end of string
		ret	
		cpi	temp,' '
		breq	pc + 2
		rcall	discard_nw
		rcall	discard_white
next_arg_end:	ret
			
;----------------------------------------------------
;---- delay temp * 10 ms (approx) 
del_hunds:	add	temp,sec_01
del_h_loop:	cpse	temp,sec_01 
		rjmp	del_h_loop
		ret
;----------------------------------------------------
;--- send a cr and lf to console
crlf:		ldi	temp,cr
		rcall	putc
		ldi	temp,lf
		rcall	putc
		ret	

;--- wait for char from console, returns in temp, and u_data
;--- will time out return a ctr-c 
getc:		ldi	temp,1
		mov	sec_01,temp
		ldi	temp,getc_to_val
getc_loop:	tst	sec_01
		brne	getc_a
		inc	sec_01
		dec	temp
		breq	getc_timeout		; timed out

getc_a:		sbrs	flags,uart_rx		; check for data
		rjmp	getc_loop
		cbr	flags,(1<<uart_rx)	; clear bit flag
		mov	temp,u_data		; move data to temp
		rcall	putc
		ret

getc_timeout:	ldi	temp,ctrl_c		;  
		mov	u_data,temp
		ret

;---- put char to console
putc:		sbis	USR,UDRE		;Is UART transmitter ready?
		rjmp	putc			;If not, wait
		out	UDR,temp		;Put character to UART
		ret

;--- convert temp to upper case
to_upper:	cpi 	temp,'a'
		brlo	toupper_end
		cpi 	temp,'z'+1
		brsh	toupper_end
		subi	temp,' '
toupper_end:	ret

;--------------------------------------
;send byte via serial as two asc chars
send_byte:	push	temp		; save temp
		rcall	send_hi_nibble	; send hi nibble
		pop	temp		; reget data	
		push	temp		; and save it again
		rcall	send_nibble
		pop	temp		; just restores temp
		ret
						; fall into send nibble
;----------------------------------------
;-send low nibble as ascii		
send_hi_nibble:	swap	temp
send_nibble:	rcall	hex_asc
		rcall	putc
		ret

;******************************************************************
;routine converts 0-f hex in temp to 0 to F ascii
;*****************************************************************
		
hex_asc:	andi   	temp,0b00001111     	;lower nybble only
   		cpi   	temp,10  		;0-9 or A-F?
   		brlo	hex_asc_1           	;
   		subi	temp,-7    		;adjust A-F
hex_asc_1:	subi   	temp,-48  		;add offset to convert to ascii
		ret

;----------------------------------------------------
;--- internal eeprom routines
;--- read eeprom (post incs EEAR!)
EERead:		cli
		sbic	EECR,EEWE	; if EEWE not clear
		rjmp	EERead		;    keep waiting

		out	EEAR,EEaddr
		sbi	EECR,EERE	; set EEPROM Read strobe
		in	EEdata,EEDR	; get data
		rjmp	inc_eear	; bump address and exit

;--- write EEPROM (post incs EEAR!)
EEWrite:	cli
		sbic	EECR,EEWE	; if EEWE not clear
		rjmp	EEWrite		;    keep waiting
 
 		out	EEAR,EEaddr
		out	EEDR,EEdata	; output the data
		sbi 	EECR,EEMWE	; set master write enable	
		sbi	EECR,EEWE	; set EEPROM Write strobe

inc_eear:	inc	EEaddr
		sei
		ret

;--- send zero terminated strings from eeprom to console
;serial_string:	out	EEAR,temp		; move temp to addr
;next_ser_char:	rcall	EERead			; get char from eeprom
;		tst	EEdata			; test for eom
;		brne	pc + 2			; skip nz
;		ret
;		mov	temp,EEdata		; 
;		rcall	putc			; ship it 
;		rjmp	next_ser_char		; loop it


;--- copy message pin and flags to registers and message data to ram :
;--- EEaddr must point to first byte of pin
get_ee_msg:	ldi	zl,at_bin+1		; point to binh
ee_msg_l1:	rcall	EEread		
		st	z+,EEdata
		cpi	zl,at_bin + 4		; done with pin?
		brne	ee_msg_l1

		rcall	EEread			; get flags
		mov	poc_flags,EEdata

		ldi	zl,ram_begin		; point to ram 
ee_msg_l2:	rcall	EEread			; copy messsage data				
		st	z+,EEdata			
		tst	EEdata			; test for end
		brne	ee_msg_l2		; keep looping
		ret

;--- copy 4 bin working registers to poc output buffer
bin2buff:	tst	poc_bit_cnt		; test to see if output reg is empty
		brne	bin2buff
		mov	poc_buff_hh,binhh	
		mov	poc_buff_h,binh
		mov	poc_buff_m,binm
		mov	poc_buff,bin
		rjmp	_2buff_end	

;--- copy 4 flash bytes to poc_buffer
flash2buff:	tst	poc_bit_cnt		; test to see if output reg is empty
		brne	flash2buff
		ldi	yl,at_poc_buff		; destination
f2p_loop:	lpm				; get byte
		st	y+,r0			;
		adiw	zl,1			; zl++
		cpi	yl,at_poc_buff + 4	; 4 bytes done?
		brne	f2p_loop

_2buff_end:	sbrc	flags,debug		; check debug flag	
		rjmp	_dump_buff
		ldi	poc_bit_cnt,32		; no debug, just send
		ret

;--- debugging, echo pocsag message to console 
_dump_buff:	push	poc_buff		; quick save data 
		push	poc_buff_m
		push	poc_buff_h 	
		push	poc_buff_hh

		ldi	poc_bit_cnt,32		; start isr 

		rcall	crlf			; make pretty
		pop	temp			
		rcall	send_byte
		pop	temp	
		rcall	send_byte
		pop	temp	
		rcall	send_byte
		pop	temp,	
		rcall	send_byte
		ret	

;--- send zero terminated strings from flash to console
; use z for pointer
ser_ram_str:	lpm				;
		tst	r0			;
		brne	pc + 2			;
		ret
		mov	temp,R0			;
		rcall	putc
		adiw	zl,1
		rjmp	ser_ram_str	
	
		
;*********************************************************************
; codespace stored  string storage 
;*********************************************************************
;--- message strings----------------
_hello:		.db	cr,lf,"AVR_POCSAG 1.0 Copyright Henry Carl Ott N2RVQ",cr,lf,0
_error:		.db	bel,cr,lf,"ERROR",0
_debug:		.db	cr,lf,"DEBUG=",0
_frame_sync:	.db	0x7C,0xD2,0x15,0xD8	;  pocsag frame sync word 
_idle: 		.db	0x7A,0x89,0xC1,0x97	;  pocsag idle word 	  
_0xaa:		.db	0xaa,0xaa,0xaa,0xaa	;  alternating bit stream for preamble	

morse_table:	.db 0b00000000		;20 space
		.db 0b10101110		;21 '!'	_._.__
		.db 0b01001010		;22 '"'	._.._.
		.db 0b00000000		;23 '#'	
		.db 0b00010011		;24 '$'	..._.._
		.db 0b00000000		;25 '%'	
		.db 0b00000000		;26 '&'	
		.db 0b01111010		;27 '''	.____.
		.db 0b10110110		;28 '(' _.__._
		.db 0b10110110		;29 ')' _.__._
		.db 0b00000000		;2A '*'
		.db 0b01010100		;2B '+' ._._.
		.db 0b11001110		;2C ',' __..__
		.db 0b10000110		;2D '-' _...._
		.db 0b01010110		;2E '.' ._._._
		.db 0b10010100		;2F '/'	_.._.


		.db 0b11111100		;30 '0' _____
		.db 0b01111100		;31 '1'	.____
		.db 0b00111100		;32 '2'	..___
		.db 0b00011100		;33 '3'	...__
		.db 0b00001100		;34 '4'	...._
		.db 0b00000100		;35 '5'	.....
		.db 0b10000100		;36 '6'	_....
		.db 0b11000100		;37 '7'	__...
		.db 0b11100100		;38 '8' ___..
		.db 0b11110100		;39 '9' ____.
		.db 0b11100010		;3A ':' ___...
		.db 0b10101010		;3B ';' _._._.
		.db 0b00000000		;3C '<'
		.db 0b10001100		;3D '=' _..._
		.db 0b00000000		;3E '>'
		.db 0b00110010		;3F '?'	..__.. 							

		.db 0b00000000		;40 '@'
		.db 0b01100000		;41 'A'	._
		.db 0b10001000		;42 'B'	_...
		.db 0b10101000		;43 'C'	_._.
		.db 0b10010000		;44 'D'	_..
		.db 0b01000000		;45 'E'	.
		.db 0b00101000		;46 'F'	.._.
		.db 0b11010000		;47 'G'	__.
		.db 0b00001000		;48 'H' ....
		.db 0b00100000		;49 'I' ..
		.db 0b01111000		;4A 'J' .___
		.db 0b10110000		;4B 'K' _._
		.db 0b01001000		;4C 'L' ._..
		.db 0b11100000		;4D 'M' __
		.db 0b10100000		;4E 'N' _.
		.db 0b11110000		;4F 'O'	___

		.db 0b01101000		;50 'P' .__.
		.db 0b11011000		;51 'Q'	__._
		.db 0b01010000		;52 'R'	._.
		.db 0b00010000		;53 'S'	...
		.db 0b11000000		;54 'T'	_
		.db 0b00110000		;55 'U'	.._
		.db 0b00011000		;56 'V'	..._
		.db 0b01110000		;57 'W'	.__
		.db 0b10011000		;58 'X' _.._
		.db 0b10111000		;59 'Y' _.__
		.db 0b11001000		;5A 'Z' __..
		.db 0b00000000		;5B '['
		.db 0b00000000		;5C '\'
		.db 0b00000000		;5D ']'
		.db 0b00000000		;5E '^'
		.db 0b00110110		;5F '_'	..__._


;*********************************************************************
; EEPROM 
;*********************************************************************
 	      	.eseg
.org		0x00
empty:		.db 	0x00			; leave empty, may get corrupted do to possible bug  
msg1_pin:	.db	0x12,0xd6,0x87
msg1_flags:	.db	(1 << m_baud_2400) | (1 << m_type_num) 
msg1_data:	.db	"212 555-1212",0x00	
.org		0x40
msg2_pin:	.db	0x12,0xd6,0x87
msg2_flags:	.db	(1 << m_fun_1) | (1<< m_fun_1) | (1 << m_type_alpha) 
msg2_data:	.db	"AVR POCSAG TEST cap=1234567",0x00	



